import java.io.*;
import java.util.*;

public class Archon {

    static public final char CBV_LAM = 1;
    static public final char CBN_LAM = 2;
    static public final char APP = 3;
    static public final char VAR = 4;
    static public final char OPEN = 5;
    static public final char VCOMP = 6;
    static public final char SWAP = 7;
    static public final char DECOMP = 8;

    // needed for useful error messages from meta-programs
    static public final char GET_LINENUM = 9;

    static public class Expr {
	char op;
	boolean normalized;
	boolean dereferenced; /* if this Expr is a var, then this is
				 true if the var has already had a
				 "cheap" dereference (without
				 cloning), and false otherwise. */
	boolean deep_clone; /* true iff this expr must be deeply cloned
			       when cloned. */
	int linenum; // unfortunately, needed for debugging
	protected Object arg1;
	protected Object arg2;
	
	protected String from; 

	public Expr(char _op, String name, Integer rename_num) {
	    assert (_op == VAR);
	    op = _op;
	    normalized = true;
	    dereferenced = false;
	    deep_clone = false;
	    linenum = 0;
	    arg1 = name;
	    arg2 = rename_num;
	    from = null;
	}

	public Expr(char _op, Expr _arg1, boolean _normalized) {
	    op = _op;
	    normalized = _normalized;
	    dereferenced = false;
	    deep_clone = false;
	    linenum = 0;
	    arg1 = _arg1;
	    arg2 = null;
	    from = null;
	}

	public Expr(char _op, Expr _arg1, Expr _arg2, boolean _normalized) {
	    op = _op;
	    normalized = _normalized;
	    dereferenced = false;
	    deep_clone = false;
	    linenum = 0;
	    arg1 = _arg1;
	    arg2 = _arg2;
	    from = null;
	}

	public Expr(char _op, Expr _arg1, Expr[] _arg2, boolean _normalized) {
	    op = _op;
	    normalized = _normalized;
	    dereferenced = false;
	    deep_clone = false;
	    linenum = 0;
	    arg1 = _arg1;
	    arg2 = _arg2;
	    from = null;
	}
    }

    public Expr app1(Expr op, Expr arg) {
	return new Expr(APP, op, arg, false);
    }

    public Expr app2(Expr op, Expr arg1, Expr arg2) {
	return new Expr(APP, app1(op, arg1), arg2, false);
    }
			

    protected int linenum;
    protected boolean debug_beta;
    protected boolean debug_beta_short;
    protected boolean debug_eval_calls;
    protected boolean debug_vars;
    protected boolean debug_assignments;
    protected boolean print_briefly;
    protected boolean print_parens;
    protected boolean print_statistics;
    protected PushbackReader stdin;
    protected boolean printing_app;
    protected int next_var;
    protected HashMap var_map; // used in parsing (see readVar())
    protected int num_beta_reductions;
    protected int num_opens;
    protected int num_vcomps;
    protected int num_swaps;
    protected int num_decompositions;
    protected int num_clones;
    protected int num_deep_clones;

    protected Expr new_var(String s) {
        return new Expr(VAR, s, new Integer(next_var++));
    }

    protected Expr e_false() {
	Expr var_true = new_var("t");
	Expr var_false = new_var("f");
	return new Expr(CBN_LAM, var_true,
			new Expr(CBN_LAM, var_false, var_false, true),
			true);
    }

    protected Expr e_true() {
	Expr var_true = new_var("t");
	Expr var_false = new_var("f");
	return new Expr(CBN_LAM, var_true,
			new Expr(CBN_LAM, var_false, var_true, true),
			true);
    }
	
    protected Expr e_none() {
	Expr var_some = new_var("s");
	Expr var_none = new_var("n");
	return new Expr(CBN_LAM, var_some,
			new Expr(CBN_LAM, var_none, var_none, true),
			true);
    }

    protected Expr e_some(Expr e) {
	Expr var_some = new_var("s");
	Expr var_none = new_var("n");
	
	return new Expr(CBN_LAM, var_some, 
			new Expr(CBN_LAM, var_none, app1(var_some, e), true),
			true);
    }

    public Archon() {
	linenum = 1;
	next_var = 1;
	debug_beta = false;
	debug_beta_short = false;
	debug_eval_calls = false;
	debug_vars = false;
	debug_assignments = false;
	print_briefly = false;
	print_parens = false;
	printing_app = false;
	stdin = new PushbackReader(new InputStreamReader(System.in), 
				   1024 /* number of bytes that can
					   be unread before we must read */);
	var_map = new HashMap();
	num_beta_reductions = 0;
	num_opens = 0;
	num_vcomps = 0;
	num_swaps = 0;
	num_decompositions = 0;
	num_clones = 0;
	num_deep_clones = 0;
    }

    protected void printStatistics(PrintStream w) {
	w.println("Opens: " + (new Integer(num_opens)));
	w.println("Vcomps: " + (new Integer(num_vcomps)));
	w.println("Swaps: " + (new Integer(num_swaps)));
	w.println("Decompositions: " + (new Integer(num_decompositions)));
	w.println("Beta reductions: " + (new Integer(num_beta_reductions)));
	w.println("Calls to clone: " + (new Integer(num_clones))
		  + " (" + (new Integer(num_deep_clones)) + " deep)");
    }

    protected Expr clone(Expr e, boolean full_expansion, 
			 boolean clone_lambdas) {
	Expr ret = null;
	char op = e.op;
	if (e.deep_clone) {
	    full_expansion = true;
	    num_deep_clones++;
	}
	switch (op) {
	case VAR: {
	    if (e.normalized) 
		return e; /* we do not want to set e's from field, so return
			     now. */

	    // Unassigned variables don't need to be cloned.  For
	    // assigned variables, the situation is complicated.  If e
	    // is assigned a variable for renaming e, then we should
	    // return that variable.  In general, we do not know at
	    // this point if a variable assigned to e was assigned by
	    // a surrounding call to clone(), or if it was assigned by
	    // beta-reduction.  For efficiency, we are trying to avoid
	    // recursively cloning things assigned by beta-reduction.
	    // But if we follow the pointer from e to something that
	    // needs to be cloned, we *must* clone it, or we will not
	    // know to clone it later.
	    // 
	    // If we are following a pointer from e to a variable, we
	    // are ok: a variable does not need to be cloned (although
	    // if it is assigned, then when its pointer is followed,
	    // the result needs to be cloned; but following that
	    // pointer will happen later).
	    //
	    // Finally, the only thing left that can go wrong is if we
	    // are in the middle of cloning a lambda abstraction whose
	    // bound variable is hidden in the term assigned to e.
	    // The only way this could have arisen is if some
	    // computation happened on that bound variable, which can
	    // only occur in an open.  We rule out that situation by
	    // always expanding all assigned variables when we come
	    // out of an open (see the code in eval below).

	    Expr next = (Expr)e.arg2;
	    
	    if (full_expansion /* uncomment to turn off optimization 3:
				  || true */) {
		boolean dereferenced = e.dereferenced;
		/* if we are already cloning lambdas, then following
		   this pointer is not a cheap dereference */ 
		e.dereferenced = dereferenced || !clone_lambdas;
		/* if we were already cloning lambdas or if e 
		   already had its cheap dereference, then we need
		   to clone lambdas */
		ret = clone(next, full_expansion, 
			    clone_lambdas || dereferenced);
		
	    }
	    else {
		if (next.op == VAR) 
		    ret=next;
		else
		    // e must have been assigned by beta-reduction, and
		    // we don't need to follow its pointer now. 
		    ret=e;
	    }
	    break;
	}

	case CBV_LAM:
	case CBN_LAM: {
	    if (clone_lambdas) {
		// we need to rename bound variables when we clone.
		// For example, consider @ \ x @ x x \ x @ x x.
		// After the first beta-reduction, we will get the
		// same term, but with the same Expr in memory for 
		// every "x".  Then when we set the pointer for that
		// "x" at the second beta-reduction, we will get a
		// mutilated term.
		//
		// So we create a new VAR Expr for the abstraction,
		// and set the old VAR Expr's second field to point
		// to it (the new VAR Expr).  When we encounter uses
		// of the old VAR, they will get replaced (in the case
		// above this one, for VAR Exprs) with uses of the
		// new variable.
		
		Expr var = (Expr)e.arg1;
		String var_name = (String)var.arg1;
		Expr new_var = new_var(var_name);
		boolean prev_normalized = var.normalized;
		Object prev = var.arg2;
		var.normalized = false; // this indicates the var is assigned.
		var.arg2 = new_var;
		Expr arg2 = (Expr)e.arg2;
		Expr e2 = clone(arg2,full_expansion,clone_lambdas);
		
		/* restore old values, so this assignment doesn't mess
		   up other copies of this term.  We are allowing this
		   dangerous situation to try to increase performance. */
		var.normalized = prev_normalized; 
		var.arg2 = prev; 
		
		ret = new Expr(op, new_var, e2, true);
	    }
	    else {
		Expr var = (Expr)e.arg1;
		Expr e2 = clone((Expr)e.arg2,full_expansion,clone_lambdas);
		ret = new Expr(op, var, e2, true);
	    }
	    break;
	}
	case DECOMP: {
	    Expr analyzed = (Expr)e.arg1;
	    Expr[] cases = (Expr[])e.arg2;
	    Expr new_analyzed = clone(analyzed,full_expansion,clone_lambdas);
	    Expr[] new_cases = new Expr[7];
	    for (int i = 0; i < 7; i++)
		new_cases[i] = clone(cases[i],full_expansion,clone_lambdas);
	    ret = new Expr(DECOMP, new_analyzed, new_cases, e.normalized);
	    break;
	}

	case SWAP: 
	case GET_LINENUM: {
	    Expr arg1 = clone((Expr)e.arg1,full_expansion,clone_lambdas);
	    ret = new Expr(op, arg1, e.normalized);
	    break;
	}

	default: {
	    Expr arg1 = clone((Expr)e.arg1,full_expansion,clone_lambdas);
	    Expr arg2 = clone((Expr)e.arg2,full_expansion,clone_lambdas);
	    ret = new Expr(op, arg1, arg2, e.normalized);
	    break;
	}
	}
	if (debug_assignments) {
	    if (ret.from == null)
		ret.from = e.from;
	}
	if (ret.linenum == 0)
	    ret.linenum = e.linenum;
	
	return ret;
    }

    protected Expr readLambda(PushbackReader s, boolean cbv)
	throws java.io.IOException {

	// read var name
	int starting_linenum = linenum;
	eat_ws(s, false);
	int var_linenum = linenum;
	String var_name = readString(s);
	eat_ws(s, false);
	
	// create var and add to var_map
	Object prev = var_map.get(var_name);
	Expr var = new_var(var_name);
	var.linenum = var_linenum;
	var_map.put(var_name, var);
	
	// read body, restore var_map
	Expr body = readExprI(s);
	var_map.put(var_name, prev);

	Expr e = new Expr(cbv ? CBV_LAM : CBN_LAM,var,body,false);
	e.linenum = starting_linenum;
	return e;
    }

    protected Expr readLet(PushbackReader s, boolean cbv)
	throws java.io.IOException {
	int starting_linenum = linenum;
	eat_ws(s, false);
	String var_name = readString(s);
	Expr var = new_var(var_name);
	
	eat_ws(s, false);
	Expr e1 = readExprI(s);

	eat_ws(s,false);
	Object prev = var_map.get(var_name);
	var_map.put(var_name, var);
	Expr e2 = readExprI(s);
	var_map.put(var_name, prev);

	Expr e = new Expr(APP, 
			  new Expr(cbv ? CBV_LAM : CBN_LAM, var, e2, true),
			  e1, false);
	e.linenum = starting_linenum;
	return e;

    }

    /* if two_args is false, we expect 1 arg */
    protected Expr readAppLike(PushbackReader s, char op, boolean two_args) 
	throws java.io.IOException {
	int starting_linenum = linenum;
	eat_ws(s, false);
	Expr arg1 = readExprI(s);
	Expr e;
	if (two_args) {
	    eat_ws(s, false);
	    e = new Expr(op,arg1,readExprI(s),false);
	}
	else 
	    e = new Expr(op,arg1,false);
	e.linenum = starting_linenum;
	return e;
    }
    protected Expr readDecomp(PushbackReader s) throws java.io.IOException {
	int starting_linenum = linenum;
	Expr[] cases = new Expr[7];
	eat_ws(s, false);
	Expr analyzed_term = readExprI(s);

	for (int i = 0; i < 7; i++) {
	    eat_ws(s, false);
	    cases[i] = readExprI(s);
	}

	Expr e = new Expr(DECOMP, analyzed_term, cases, false);
	e.linenum = starting_linenum;
	return e;
    }

    protected Expr readVarUse(PushbackReader s) 
	throws java.io.IOException 
    {
	String name = readString(s);
	Expr var = (Expr)var_map.get(name);
	if (var == null) {
	    var = new Expr(VAR, name, new Integer(0));
	    var.linenum = linenum;
	}
	return var;
    }

    protected String readString(PushbackReader s) 
	throws java.io.IOException {
	int i;
	StringBuffer b = new StringBuffer();
	while ((i = s.read()) != -1) {
	    char c = (char)i;
	    if (Character.isWhitespace(c) || is_operator_char(c)) {
		s.unread(i);
		break;
	    }
	    else
		b.append(c);
	}
	if (b.length() == 0) 
	    parse_error("");
	return b.toString();
    }
	
    protected boolean next_is_ws(PushbackReader s) 
	throws java.io.IOException {
	int i = s.read();
	if (i == -1) 
	    return false;
	s.unread(i);
	return (Character.isWhitespace((char)i));
    }

    /* return an Expr holding the whitespace read iff some was read
       and ret_expr is true. */
    protected Expr eat_ws(PushbackReader s, boolean ret_expr)
	throws java.io.IOException {
	int i;
	int starting_linenum = linenum;
	StringBuffer b = new StringBuffer();
	while ((i = s.read()) != -1) {
	    char c = (char) i;
	    int cmtCnt = 0;
	    if (!Character.isWhitespace(c)) {
		if (c == '('){
		    int i0 = i;
		    if((c = (char) (i = s.read())) != '*'){
			s.unread(i);
			s.unread(i0);
			break; // out of while loop
		    }
		    else {
			cmtCnt++;
			while (cmtCnt > 0) {
			    i = s.read();
			    c = (char) i;
			    if (c == '('){
			        if((c = (char) (i = s.read())) != '*')
				    {s.unread(i);}
				else{cmtCnt++;}
			    }
			    else if (c == '*'){
				if((c = (char) (i = s.read())) != ')')
				    {s.unread(i);}
				else{
				    cmtCnt--; 
				}
			    }
			    else if (c == '\n')
				linenum++;
			    if(cmtCnt < 0){
				System.out.println
				    ("ERROR: nested comment broke parsing!");
				System.exit(1);
			    }
			}
		    }
		}
		else if (c == '%'){
		    // read a comment
		    while ((i = s.read()) != -1 && i != '\n');
		    linenum++;
		}
		else {
		    s.unread(i);
		    break;
		}
	    }
	    else {
		b.append(c);
		if (c == '\n')
		    linenum++;
	    }
	}
	if (b.length() == 0)
	    return null;
	Expr e = new Expr(VAR, b.toString(), new Integer(0));
	e.linenum = starting_linenum;
	return e;
    }

    protected void parse_error(String s) {
	System.out.print("Parse error on line "+
			 (new Integer(linenum)).toString());
	if (s.equals(""))
	    System.out.println(".");
	else {
	    System.out.println(":");
	    System.out.println(s);
	}
	System.exit(1);
    }

    protected boolean is_operator_char(char c) {
	switch (c) {
	case '\\':
	case '^':
	case '@':
	case '!':
	case '?':
	case '#':
	case '%':
	case '*':
	case '(':
	case ')':
	case '=':
	case ':':
	case '$':
	case '&':
	case '`':
	case '\'':
	    return true;
	}
	return false;
    }

    protected Expr readExprI(PushbackReader s)
	throws java.io.IOException {
	int i = s.read();
	if (i == -1) 
	    parse_error("");
	char c = (char)i;
	Expr ret;
	switch (c) {
	case '\\':
	    ret = readLambda(s,true);
	    break;
	case '^':
	    ret = readLambda(s,false);
	    break;
	case '@':
	    ret = readAppLike(s,APP,true);
	    break;
	case ')':
	    parse_error("Unexpected close parenthesis on line "+
			(new Integer(linenum)));
	    ret = null;
	    break;
	case '(': {
	    int starting = linenum;
	    eat_ws(s,false);
	    ret = readExprI(s);
	    eat_ws(s, false);
	    if ((char)s.read() != ')')
		parse_error("Missing close parenthesis for open parenthesis on"
			    +" line "+(new Integer(starting)));
	    break;
	}
        case '$': 
	    ret = readLet(s,true);
	    break;
        case '&': 
	    ret = readLet(s,false);
	    break;
	case '!':
	    ret = readAppLike(s,OPEN,true);
	    break;
	case '#':
	    ret = readAppLike(s,SWAP,false);
	    break;
	case '=':
	    ret = readAppLike(s,VCOMP,true);
	    break;
	case ':':
	    ret = readDecomp(s);
	    break;
	case '?':
	    ret = readAppLike(s,GET_LINENUM,false);
	    break;
	case '\'':
	    ret = readVarUse(s);
	    break;
	default:
	    s.unread(i);
	    ret = readVarUse(s);
	    break;
	}
	return ret;
    }

    public Expr readExpr(PushbackReader s) throws java.io.IOException {
	linenum = 1;
	eat_ws(s, false);
	Expr e = readExprI(s);
	eat_ws(s, false);
	if (s.read() != -1)
	    parse_error("");
	return e;
    }

    protected void print_binary(String op, Expr e1, Expr e2, 
				HashMap in_scope, PrintStream w) 
	throws java.io.IOException {
	w.print(op);
	w.print(' ');
	printExpr(e1, in_scope, w);
	w.print(' ');
	printExpr(e2, in_scope, w);
    }

    public void printVarName(String var_name, PrintStream w) 
	throws java.io.IOException {
	if (is_operator_char(var_name.charAt(0)))
	    w.print('\'');
	w.print(var_name);
    }
		    

    public void printExpr(Expr e, HashMap in_scope,
			  PrintStream w) throws java.io.IOException {
	char op = e.op;
	boolean printing_app_prev = printing_app;
	boolean print_from = (debug_assignments && (e.from != null) &&
			      (op != VAR) && 
			      /* the next case prevents double printing
				 of from information when we are printing
				 parens */ 
			      !(op == APP && print_parens &&
				!printing_app_prev));
	if (debug_assignments && e.deep_clone)
	    w.print("`");
	if (print_from) {
	    w.print(e.from);
            w.print("[");
	}
	if (print_from && print_briefly) {
	    // skip printing e
	}
	else {

	    // by default, we are not printing an application
	    printing_app = false;

	    switch(op) {
	    case CBV_LAM:
	    case CBN_LAM: {

		/* the idea here is to maintain a mapping from each
		   variable name to its de Bruijn level, and to map
		   each variable (using the assignment mechanism, via
		   the variable's arg2 field) to a new variable which
		   has that de Bruijn level as its numeric part (its
		   arg2 value). */

		Expr var = (Expr)e.arg1;
		Object var_name = var.arg1;
		Object prev_val = var.arg2;
		boolean prev_normalized = var.normalized;
		Integer prev = (Integer)in_scope.get(var_name);
		Integer count = prev;
		if (count == null)
		    count = new Integer(0);
		else
		    count = new Integer(count.intValue() + 1);

		if (op == CBV_LAM)
		    w.print("\\ ");
		else
		    w.print("^ ");
		in_scope.put(var_name, count);
		var.arg2 = new Expr(VAR, (String) var_name, count);
		var.normalized = false;
		printExpr((Expr)e.arg1, in_scope, w);
		w.print(' ');
		printExpr((Expr)e.arg2, in_scope, w);
		in_scope.put(var_name, prev);
		var.normalized = prev_normalized;
		var.arg2 = prev_val;
		break;
	    }
	    case VAR: {
		if (e.normalized) {
		    String var_name = (String)e.arg1;  
		    printVarName(var_name,w);
		    Integer count = (Integer)e.arg2;
		    if (debug_vars)
			w.print(count);
		    else if (count != null && count.intValue() > 0) { 
			w.print("\'");
			w.print(count);
		    }
		}
		else {
		    // this variable has a value assigned by beta-reduction
		    if (debug_assignments) {
			w.print("*");
			if (e.dereferenced)
			    w.print("~");
		    }
		    printExpr((Expr)e.arg2, in_scope, w);
		}
		    
		break;
	    }
	    case APP: {
		Expr e1 = (Expr)e.arg1;
		if (printing_app_prev || !print_parens) {
		    w.print("@ ");
		    printing_app = true;
		    printExpr((Expr)e.arg1, in_scope, w);
		    w.print(' ');
		    printing_app = false;
		    printExpr((Expr)e.arg2, in_scope, w);
		}
		else {
		    w.print('(');
		    printing_app = true;
		    printExpr(e, in_scope, w);
		    w.print(')');
		}
		break;
	    }
	    case VCOMP:
		print_binary("=",(Expr)e.arg1, (Expr)e.arg2, in_scope, w);
		break;
	    case SWAP:
		w.print("# ");
		printExpr((Expr)e.arg1, in_scope, w);
		break;
	    case GET_LINENUM:
		w.print("? ");
		printExpr((Expr)e.arg1, in_scope, w);
		break;
	    case OPEN:
		print_binary("!",(Expr)e.arg1, (Expr)e.arg2, in_scope, w);
		break;
	    case DECOMP:
		w.print(": ");
		printExpr((Expr)e.arg1, in_scope, w);
		Expr[] cases = (Expr[])e.arg2;
		for (int i=0; i < 7; i++) {
		    w.print(' ');
		    printExpr(cases[i], in_scope, w);
		}
		break;

	    }

	    // restore printing_app
	    printing_app = printing_app_prev;
		
	}
	if (print_from)
	    w.print("]");
    }

    protected Expr returning(Expr e) throws java.io.IOException {
	if (debug_eval_calls) {
	    System.out.println("--------------------------------");
	    System.out.print("RETURNING ");
	    printExpr(e, new HashMap(), System.out);
	    System.out.println(")");
	}
	return e;
    }


    // if e is a variable, follow its pointer and repeat; clone the
    // result.  Otherwise, return e.
    protected Expr expand_if_var(Expr e, boolean do_clone) {
	boolean changed = false;
	Expr orig = e;
	Expr last = orig;
	while (e.op == VAR && !e.normalized) {
	    changed = true;
	    last = e;
	    e = (Expr)e.arg2;
	}
	if (changed) {
	    if (debug_assignments)
		e.from = (String)orig.arg1;
	    if (do_clone) {
		if (last.dereferenced
		    /* uncomment to turn off optimization 1: 
		       || true*/) {
		    num_clones++;
		    return clone(e, false /* full expansion */,
				 true /* clone lambdas */);
		}
		else {
		    last.dereferenced = true;
		    return e;
		}
	    }
	}
	return e;
    }

    public Expr eval(Expr _e) throws java.io.IOException 
    {
	Expr e = _e;

	if (debug_eval_calls) {
	    System.out.println("--------------------------------");
	    System.out.print("( EVALUATING ");
	    printExpr(e, new HashMap(), System.out);
	    System.out.println();
	}

	while (true) {
	    if (e.normalized)
		// this includes the case for free variables
		return returning(e);

	    char op = e.op;
	    switch (op) {
	    case CBV_LAM:
	    case CBN_LAM: {
		return returning(e);
	    } 
	    case VAR: {
		// any non-normalized variable has second pointer
		// set to a value assigned to that variable by beta-reduction.
		e = expand_if_var(e,true /* we must clone here, because
					    we might end up applying the
					    new e. */);
		break; // out of switch
	    }
	    case OPEN: {
		num_opens++;
		Expr e1 = expand_if_var((Expr)e.arg1, 
					true /* we need to clone now */);
		char e1op = e1.op;
		if (e1op != CBV_LAM && e1op != CBN_LAM) {
		    Expr ret = e_false();
		    return returning(new Expr(CBV_LAM, new_var("z"), ret, 
					      true));
		}
		Expr var = (Expr)e1.arg1;
		Expr e2 = (Expr)e.arg2;
		Expr ret = new Expr(e1op, var, 
				    eval(app2(e2, var,
					      (Expr)e1.arg2)),
				    true);
		ret.linenum = e2.linenum;

		/* if we do need to clone ret, we must deeply clone it
		   (clone with full expansion).  This is because
		   otherwise we can leave the bound variable trapped
		   in a value assigned to another variable.  That case
		   is problematic for clone() above: trapped variables
		   can then escape their scope during cloning. */
		ret.deep_clone = true;
		return returning(ret);
	    }
	    case VCOMP: {
		num_vcomps++;
		Expr e1 = expand_if_var((Expr)e.arg1,false);
		if (e1.op != VAR)
		    return returning(e_false());
		Expr e2 = expand_if_var((Expr)e.arg2,false);
		if (e2.op != VAR)
		    return returning(e_false());
		if (e1.arg1.equals(e2.arg1) &&
		    e1.arg2.equals(e2.arg2))
		    // e1 and e2 are equal variables
		    return returning(e_true());
		return returning(e_false());
	    }
	    case SWAP: {
		num_swaps++;
		Expr e1 = eval((Expr)e.arg1);
		char e1op = e1.op;
		if (e1op != CBV_LAM && e1op != CBN_LAM) {
		    Expr ret = e_false();
		    return returning(new Expr(CBV_LAM, new_var("z"), ret, 
					      true));
		}
		Expr e1_body = (Expr)e1.arg2;
		char e1_body_op = e1_body.op;
		if (e1_body_op != CBV_LAM && e1_body_op != CBN_LAM) {
		    Expr ret = e_false();
		    return returning(new Expr(CBV_LAM, new_var("z"), ret, 
					      true));
		}

		Expr ret_body = new Expr(e1op, (Expr)e1.arg1, 
					 (Expr)e1_body.arg2, true);
		Expr ret = new Expr(e1_body_op, (Expr)e1_body.arg1,
				    ret_body,
				    true);
		ret.linenum = e1.linenum;
		// swap the deep cloning information
		ret.deep_clone = e1_body.deep_clone;
		ret_body.deep_clone = e1.deep_clone;
		return returning(ret); 
	    }
	    case DECOMP: {
		num_decompositions++;
		Expr new_analyzed = expand_if_var((Expr)e.arg1,true);
		Expr[] cases = (Expr[])e.arg2;
		switch (new_analyzed.op) {
		case VAR:
		    e = app1(cases[0],new_analyzed);
		    break;
		case CBV_LAM:
		    e = app2(cases[1], e_true(), new_analyzed);
		    break;
		case CBN_LAM:
		    e = app2(cases[1], e_false(), new_analyzed);
		    break;
		case APP:
		    e = app2(cases[2], (Expr)new_analyzed.arg1,
			     (Expr)new_analyzed.arg2);
		    break;
		case OPEN:
		    e = app2(cases[3],
			     (Expr)new_analyzed.arg1,
			     (Expr)new_analyzed.arg2);
		    break;
		case VCOMP:
		    e = app2(cases[4],
			     (Expr)new_analyzed.arg1,
			     (Expr)new_analyzed.arg2);
		    break;
		case SWAP:
		    e = app1(cases[5],
			     (Expr)new_analyzed.arg1);
		    break;
		case DECOMP: {
		    e = app1(cases[6], (Expr)new_analyzed.arg1);
		    Expr[] new_analyzed_cases = (Expr[])new_analyzed.arg2;
		    for (int i = 0; i < 7; i++)
			e = app1(e, new_analyzed_cases[i]);
		    break;
		}
		}
	    }
	    case APP: {
		Expr e1 = eval((Expr)e.arg1);
		Expr e2 = (Expr)e.arg2;
		char e1op = e1.op;
		if (e1op != CBN_LAM)
		    e2 = eval(e2);
		if (e1op == CBV_LAM || e1op == CBN_LAM) {
		    num_beta_reductions++;
		    Expr var = (Expr)e1.arg1;
		    e = (Expr)e1.arg2; // var's pointer set just below

		    if (debug_beta) {
			System.out.println("--------------------------------");
			System.out.println("BETA:");
			printExpr(var, new HashMap(), System.out);
			if (debug_beta_short) 
			    System.out.println();
			else {
			    System.out.print(" = ");
			    printExpr(e2, new HashMap(), System.out);
			    System.out.println();
			    printExpr(e, new HashMap(), System.out);
			    System.out.println();
			}
		    }

		    assert (var.normalized);

		    if (debug_assignments)
			e2.from = (String)var.arg1;
		    var.arg2 = e2;
		    var.normalized = false;

		    break; // out of switch
                }

		Expr ret = new Expr(APP, e1, e2, true);
		ret.linenum = e.linenum;
		return returning(ret);
	    }

	    case GET_LINENUM: {
		Expr e1 = expand_if_var((Expr)e.arg1,false);
		return returning(new Expr(VAR,
					  "line_"+(new Integer(e1.linenum)),
					  new Integer(0)));
	    }
		

	    } // switch
	} // while
    }

    static public void main(String args[]) {
	String input_file = null;
	Archon archon = new Archon();
	boolean no_eval = false;
	boolean no_output = false;
	boolean loop_at_end = false;
	for (int i = 0; i < args.length; i++) 
	    if (args[i].equals("--help")) {
		System.out.println("Archon interpreter.");
		System.out.println("Debugging options:");
		System.out.println("  --debug");
		System.out.println("  --debug-assignments");
		System.out.println("  --debug-beta");
		System.out.println("  --debug-beta-short");
		System.out.println("  --debug-eval-calls");
		System.out.println("");
		System.out.println("Other options:");
		System.out.println("  --print-brackets");
		System.out.println("  --print-statistics");
		System.out.println("  --print-briefly");
		System.out.println("  --no-eval");
		System.out.println("  --loop-at-end");
		System.out.println("  --no-output");
		System.out.println("  --help");
	    }
	    else if (args[i].equals("--debug-beta"))
		archon.debug_beta = true;
	    else if (args[i].equals("--debug-beta-short")) {
		archon.debug_beta = true;
		archon.debug_beta_short = true;
	    }
	    else if (args[i].equals("--debug-eval-calls"))
		archon.debug_eval_calls = true;
	    else if (args[i].equals("--debug-vars"))
		archon.debug_vars = true;
	    else if (args[i].equals("--debug")) {
		archon.debug_assignments = true;
		archon.print_parens = true;
	    }
	    else if (args[i].equals("--debug-assignments"))
		archon.debug_assignments = true;
	    else if (args[i].equals("--print-statistics"))
		archon.print_statistics = true;
	    else if (args[i].equals("--print-parens"))
		archon.print_parens = true;
	    else if (args[i].equals("--print-briefly")) {
		archon.print_briefly = true;
		archon.debug_assignments = true;
	    }
	    else if (args[i].equals("--debug-all")) {
		archon.debug_vars = true;
		archon.debug_beta = true;
		archon.debug_assignments = true;
		archon.debug_eval_calls = true;
	    }
	    else if (args[i].equals("--no-eval")) 
		no_eval = true;
	    else if (args[i].equals("--no-output")) 
		no_output = true;
	    else if (args[i].equals("--loop-at-end")) 
		loop_at_end = true;
	    else
		input_file = args[i];

	if (input_file == null) {
	    System.out.println("Archon expects the name of one file"+
			       " to execute.");
	    System.exit(1);
	}
	PushbackReader s = null;
	try {
	    s = new PushbackReader(new FileReader(input_file),2);
	}
	catch (java.io.FileNotFoundException e) {
	    System.out.println("Archon could not open \""+args[0]+"\" for"+
			       " reading.");
	    System.exit(1);
	}

	try {
	    Expr e = archon.readExpr(s);
	    Expr r = no_eval ? e : archon.eval(e);
	    PrintStream w = 
		new PrintStream(new BufferedOutputStream(System.out));
	    if (!no_output) 
		archon.printExpr(r, new HashMap(), w);
	    if (archon.print_statistics) {
		PrintStream w2 = 
		    new PrintStream(new BufferedOutputStream(System.err));
		w2.println("");
		archon.printStatistics(w2);
		w2.flush();
	    }
	    w.flush();
	    System.out.println();
	}
	catch (java.io.IOException e) {
	    System.out.println("Caught IOException: " + e);
	    System.exit(1);
	}   
	if (loop_at_end)
	    while(true) ;
    }
}
